@@ -1,22 +1,10 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
  * Copyright (c) 2000-2001,2005 Silicon Graphics, Inc.
  * All Rights Reserved.
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation.
- *
- * This program is distributed in the hope that it would be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write the Free Software Foundation,
- * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <libxfs.h>
+#include "libxfs.h"
 #include "globals.h"
 #include "agheader.h"
 #include "protos.h"
@@ -30,7 +18,7 @@
 static int
 verify_set_agf(xfs_mount_t *mp, xfs_agf_t *agf, xfs_agnumber_t i)
 {
-	xfs_drfsbno_t agblocks;
+	xfs_rfsblock_t agblocks;
 	int retval = 0;
 
 	/* check common fields */
@@ -69,11 +57,11 @@ verify_set_agf(xfs_mount_t *mp, xfs_agf_t *agf, xfs_agnumber_t i)
 				be32_to_cpu(agf->agf_length), i,
 				mp->m_sb.sb_agblocks);
 			if (!no_modify)
-				agf->agf_length = 
+				agf->agf_length =
 					cpu_to_be32(mp->m_sb.sb_agblocks);
 		} else  {
 			agblocks = mp->m_sb.sb_dblocks -
-				(xfs_drfsbno_t) mp->m_sb.sb_agblocks * i;
+				(xfs_rfsblock_t) mp->m_sb.sb_agblocks * i;
 
 			if (be32_to_cpu(agf->agf_length) != agblocks)  {
 				retval = XR_AG_AGF;
@@ -91,18 +79,18 @@ verify_set_agf(xfs_mount_t *mp, xfs_agf_t *agf, xfs_agnumber_t i)
 	 * check first/last AGF fields.  if need be, lose the free
 	 * space in the AGFL, we'll reclaim it later.
 	 */
-	if (be32_to_cpu(agf->agf_flfirst) >= XFS_AGFL_SIZE(mp))  {
-		do_warn(_("flfirst %d in agf %d too large (max = %zu)\n"),
+	if (be32_to_cpu(agf->agf_flfirst) >= libxfs_agfl_size(mp)) {
+		do_warn(_("flfirst %d in agf %d too large (max = %u)\n"),
 			be32_to_cpu(agf->agf_flfirst),
-			i, XFS_AGFL_SIZE(mp));
+			i, libxfs_agfl_size(mp) - 1);
 		if (!no_modify)
 			agf->agf_flfirst = cpu_to_be32(0);
 	}
 
-	if (be32_to_cpu(agf->agf_fllast) >= XFS_AGFL_SIZE(mp))  {
-		do_warn(_("fllast %d in agf %d too large (max = %zu)\n"),
+	if (be32_to_cpu(agf->agf_fllast) >= libxfs_agfl_size(mp)) {
+		do_warn(_("fllast %d in agf %d too large (max = %u)\n"),
 			be32_to_cpu(agf->agf_fllast),
-			i, XFS_AGFL_SIZE(mp));
+			i, libxfs_agfl_size(mp) - 1);
 		if (!no_modify)
 			agf->agf_fllast = cpu_to_be32(0);
 	}
@@ -112,7 +100,7 @@ verify_set_agf(xfs_mount_t *mp, xfs_agf_t *agf, xfs_agnumber_t i)
 	if (!xfs_sb_version_hascrc(&mp->m_sb))
 		return retval;
 
-	if (platform_uuid_compare(&agf->agf_uuid, &mp->m_sb.sb_uuid)) {
+	if (platform_uuid_compare(&agf->agf_uuid, &mp->m_sb.sb_meta_uuid)) {
 		char uu[64];
 
 		retval = XR_AG_AGF;
@@ -120,7 +108,8 @@ verify_set_agf(xfs_mount_t *mp, xfs_agf_t *agf, xfs_agnumber_t i)
 		do_warn(_("bad uuid %s for agf %d\n"), uu, i);
 
 		if (!no_modify)
-			platform_uuid_copy(&agf->agf_uuid, &mp->m_sb.sb_uuid);
+			platform_uuid_copy(&agf->agf_uuid,
+					   &mp->m_sb.sb_meta_uuid);
 	}
 	return retval;
 }
@@ -128,7 +117,7 @@ verify_set_agf(xfs_mount_t *mp, xfs_agf_t *agf, xfs_agnumber_t i)
 static int
 verify_set_agi(xfs_mount_t *mp, xfs_agi_t *agi, xfs_agnumber_t agno)
 {
-	xfs_drfsbno_t agblocks;
+	xfs_rfsblock_t agblocks;
 	int retval = 0;
 
 	/* check common fields */
@@ -167,11 +156,11 @@ verify_set_agi(xfs_mount_t *mp, xfs_agi_t *agi, xfs_agnumber_t agno)
 				be32_to_cpu(agi->agi_length), agno,
 					mp->m_sb.sb_agblocks);
 			if (!no_modify)
-				agi->agi_length = 
+				agi->agi_length =
 					cpu_to_be32(mp->m_sb.sb_agblocks);
 		} else  {
 			agblocks = mp->m_sb.sb_dblocks -
-				(xfs_drfsbno_t) mp->m_sb.sb_agblocks * agno;
+				(xfs_rfsblock_t) mp->m_sb.sb_agblocks * agno;
 
 			if (be32_to_cpu(agi->agi_length) != agblocks)  {
 				retval = XR_AG_AGI;
@@ -190,7 +179,7 @@ verify_set_agi(xfs_mount_t *mp, xfs_agi_t *agi, xfs_agnumber_t agno)
 	if (!xfs_sb_version_hascrc(&mp->m_sb))
 		return retval;
 
-	if (platform_uuid_compare(&agi->agi_uuid, &mp->m_sb.sb_uuid)) {
+	if (platform_uuid_compare(&agi->agi_uuid, &mp->m_sb.sb_meta_uuid)) {
 		char uu[64];
 
 		retval = XR_AG_AGI;
@@ -198,7 +187,8 @@ verify_set_agi(xfs_mount_t *mp, xfs_agi_t *agi, xfs_agnumber_t agno)
 		do_warn(_("bad uuid %s for agi %d\n"), uu, agno);
 
 		if (!no_modify)
-			platform_uuid_copy(&agi->agi_uuid, &mp->m_sb.sb_uuid);
+			platform_uuid_copy(&agi->agi_uuid,
+					   &mp->m_sb.sb_meta_uuid);
 	}
 
 	return retval;
@@ -245,13 +235,18 @@ compare_sb(xfs_mount_t *mp, xfs_sb_t *sb)
  * superblocks, not just the secondary superblocks.
  */
 static int
-secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
-	xfs_agnumber_t i)
+secondary_sb_whack(
+	struct xfs_mount *mp,
+	struct xfs_buf	*sbuf,
+	struct xfs_sb	*sb,
+	xfs_agnumber_t	i)
 {
-	int do_bzero;
-	int size;
-	char *ip;
-	int rval;
+	struct xfs_dsb	*dsb = XFS_BUF_TO_SBP(sbuf);
+	int		do_bzero = 0;
+	int		size;
+	char		*ip;
+	int		rval = 0;;
+	uuid_t		tmpuuid;
 
 	rval = do_bzero = 0;
 
@@ -263,7 +258,10 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	 *
 	 * size is the size of data which is valid for this sb.
 	 */
-	if (xfs_sb_version_hascrc(sb))
+	if (xfs_sb_version_hasmetauuid(sb))
+		size = offsetof(xfs_sb_t, sb_meta_uuid)
+			+ sizeof(sb->sb_meta_uuid);
+	else if (xfs_sb_version_hascrc(sb))
 		size = offsetof(xfs_sb_t, sb_lsn)
 			+ sizeof(sb->sb_lsn);
 	else if (xfs_sb_version_hasmorebits(sb))
@@ -275,16 +273,13 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	else if (xfs_sb_version_hassector(sb))
 		size = offsetof(xfs_sb_t, sb_logsectsize)
 			+ sizeof(sb->sb_logsectsize);
-	else if (xfs_sb_version_hasdirv2(sb))
+	else /* only support dirv2 or more recent */
 		size = offsetof(xfs_sb_t, sb_dirblklog)
 			+ sizeof(sb->sb_dirblklog);
-	else
-		size = offsetof(xfs_sb_t, sb_width)
-			+ sizeof(sb->sb_width);
 
 	/* Check the buffer we read from disk for garbage outside size */
-	for (ip = XFS_BUF_PTR(sbuf) + size;
-	     ip < XFS_BUF_PTR(sbuf) + mp->m_sb.sb_sectsize;
+	for (ip = (char *)sbuf->b_addr + size;
+	     ip < (char *)sbuf->b_addr + mp->m_sb.sb_sectsize;
 	     ip++)  {
 		if (*ip)  {
 			do_bzero = 1;
@@ -304,10 +299,13 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 			 * be set, and the latter is never updated past
 			 * the last field - just zap them both.
 			 */
-			memset((void *)((__psint_t)sb + size), 0,
+			memcpy(&tmpuuid, &sb->sb_meta_uuid, sizeof(uuid_t));
+			memset((void *)((intptr_t)sb + size), 0,
 				mp->m_sb.sb_sectsize - size);
-			memset(XFS_BUF_PTR(sbuf) + size, 0,
+			memset((char *)sbuf->b_addr + size, 0,
 				mp->m_sb.sb_sectsize - size);
+			/* Preserve meta_uuid so we don't fail uuid checks */
+			memcpy(&sb->sb_meta_uuid, &tmpuuid, sizeof(uuid_t));
 		} else
 			do_warn(
 	_("would zero unused portion of %s superblock (AG #%u)\n"),
@@ -326,7 +324,7 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	if (sb->sb_flags)  {
 		if (!no_modify)
 			sb->sb_flags = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(_("bad flags field in superblock %d\n"), i);
 		} else
@@ -334,15 +332,21 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	}
 
 	/*
-	 * quota inodes and flags in secondary superblocks
-	 * are never set by mkfs.  However, they could be set
-	 * in a secondary if a fs with quotas was growfs'ed since
-	 * growfs copies the new primary into the secondaries.
+	 * quota inodes and flags in secondary superblocks are never set by
+	 * mkfs.  However, they could be set in a secondary if a fs with quotas
+	 * was growfs'ed since growfs copies the new primary into the
+	 * secondaries.
+	 *
+	 * Also, the in-core inode flags now have different meaning to the
+	 * on-disk flags, and so libxfs_sb_to_disk cannot directly write the
+	 * sb_gquotino/sb_pquotino fields without specific sb_qflags being set.
+	 * Hence we need to zero those fields directly in the sb buffer here.
 	 */
-	if (sb->sb_inprogress == 1 && sb->sb_uquotino)  {
+
+	if (sb->sb_inprogress == 1 && sb->sb_uquotino != NULLFSINO)  {
 		if (!no_modify)
 			sb->sb_uquotino = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(
 		_("non-null user quota inode field in superblock %d\n"),
@@ -352,10 +356,12 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 			rval |= XR_AG_SB_SEC;
 	}
 
-	if (sb->sb_inprogress == 1 && sb->sb_gquotino)  {
-		if (!no_modify)
+	if (sb->sb_inprogress == 1 && sb->sb_gquotino != NULLFSINO)  {
+		if (!no_modify) {
 			sb->sb_gquotino = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
+			dsb->sb_gquotino = 0;
+		}
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(
 		_("non-null group quota inode field in superblock %d\n"),
@@ -365,10 +371,18 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 			rval |= XR_AG_SB_SEC;
 	}
 
-	if (sb->sb_inprogress == 1 && sb->sb_pquotino)  {
-		if (!no_modify)
+	/*
+	 * Note that sb_pquotino is not considered a valid sb field for pre-v5
+	 * superblocks. If it is anything other than 0 it is considered garbage
+	 * data beyond the valid sb and explicitly zeroed above.
+	 */
+	if (xfs_sb_version_has_pquotino(&mp->m_sb) &&
+	    sb->sb_inprogress == 1 && sb->sb_pquotino != NULLFSINO)  {
+		if (!no_modify) {
 			sb->sb_pquotino = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
+			dsb->sb_pquotino = 0;
+		}
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(
 		_("non-null project quota inode field in superblock %d\n"),
@@ -381,7 +395,7 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	if (sb->sb_inprogress == 1 && sb->sb_qflags)  {
 		if (!no_modify)
 			sb->sb_qflags = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(_("non-null quota flags in superblock %d\n"),
 				i);
@@ -395,22 +409,10 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	 * written at mkfs time (and the corresponding sb version bits
 	 * are set).
 	 */
-	if (!xfs_sb_version_hasshared(sb) && sb->sb_shared_vn != 0)  {
-		if (!no_modify)
-			sb->sb_shared_vn = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
-			rval |= XR_AG_SB;
-			do_warn(
-		_("bad shared version number in superblock %d\n"),
-				i);
-		} else
-			rval |= XR_AG_SB_SEC;
-	}
-
 	if (!xfs_sb_version_hasalign(sb) && sb->sb_inoalignmt != 0)  {
 		if (!no_modify)
 			sb->sb_inoalignmt = 0;
-		if (sb->sb_versionnum & XR_PART_SECSB_VNMASK || !do_bzero)  {
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(
 		_("bad inode alignment field in superblock %d\n"),
@@ -423,7 +425,7 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 	    (sb->sb_unit != 0 || sb->sb_width != 0))  {
 		if (!no_modify)
 			sb->sb_unit = sb->sb_width = 0;
-		if (sb->sb_versionnum & XR_GOOD_SECSB_VNMASK || !do_bzero)  {
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(
 		_("bad stripe unit/width fields in superblock %d\n"),
@@ -441,7 +443,7 @@ secondary_sb_wack(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 			sb->sb_logsectsize = 0;
 			sb->sb_logsectlog = 0;
 		}
-		if (sb->sb_versionnum & XR_GOOD_SECSB_VNMASK || !do_bzero)  {
+		if (!do_bzero)  {
 			rval |= XR_AG_SB;
 			do_warn(
 		_("bad log/data device sector size fields in superblock %d\n"),
@@ -506,7 +508,7 @@ verify_set_agheader(xfs_mount_t *mp, xfs_buf_t *sbuf, xfs_sb_t *sb,
 		rval |= XR_AG_SB;
 	}
 
-	rval |= secondary_sb_wack(mp, sbuf, sb, i);
+	rval |= secondary_sb_whack(mp, sbuf, sb, i);
 
 	rval |= verify_set_agf(mp, agf, i);
 	rval |= verify_set_agi(mp, agi, i);
